/* Part of the SUDK sudoku puzzle solver/editor */
/* By Alex Wilson, copyright 2006 */

/* Solving algorithm based in part on the freeware work of
   Wim Colgate */

#include "solving.h"

DEFINE_INT_HANDLER (kill_loop) {
	kill_search = 2;
}

int place_next(int loc, sudoku *pboard) {
	while ((loc < ARRAY_SIZE) && (sudoku_get(pboard,loc) != 0)) loc++;
	if (loc < ARRAY_SIZE) return loc;
	return -1;
}

int row_check(int loc, int val, sudoku *pboard) {
	int i = BASE(loc);
	int j;

	for (j=0; j<ROW_SIZE; j++) {
		if (sudoku_get(pboard,i+j) == val) return 0;
	}

	return 1;
}

int column_check(int loc, int val, sudoku *pboard) {
	int i = TOP(loc);
	int j;

	for (j=0; j<ROW_SIZE; j++) {
		if (sudoku_get(pboard,i+j*ROW_SIZE) == val) return 0;
	}

	return 1;
}

int square_check(int loc, int val, sudoku *pboard) {
	int i = ULEFT(loc);
	int j,k;

	for (j=0; j<SQUARE_Y; j++) {
		for (k=0; k<SQUARE_X; k++) {
			if (sudoku_get(pboard,i+j*ROW_SIZE+k) == val) return 0;
		}
	}

	return 1;
}

int sudo(int loc, int num, sudoku *board) {
	sudoku *nboard;

	if (num > ROW_SIZE) return -1;
	if (kill_search != 0) return -1;

	loc = place_next(loc, board);

	if (loc == -1) {
		memcpy(solution, board, sizeof(sudoku));
		kill_search = 1;
		return -1;
	}

	if (row_check(loc,num,board) && column_check(loc,num,board) && square_check(loc,num,board)) {
		nboard = malloc(sizeof(sudoku));
		if (nboard == NULL) return -1;
		memcpy(nboard, board, sizeof(sudoku));

		sudoku_set(nboard,loc,num);
		sudo(loc+1, 1, nboard);
		free(nboard);
	}

	if (kill_search == 0) sudo(loc, num+1, board);
}

int solve(sudoku *board) {
	INT_HANDLER OldInt = NULL;

	sol_handle = HeapAllocHigh(sizeof(sudoku));
	if (sol_handle == H_NULL) return 0;
	solution = HeapDeref(sol_handle);

	memcpy(solution, board, sizeof(sudoku));

	kill_search = 0;

	OldInt = GetIntVec(INT_VEC_ON_KEY_PRESS);
	SetIntVec(INT_VEC_ON_KEY_PRESS,kill_loop);

	TRY
		sudo(0, 1, board);
	ONERR
		HeapFree(sol_handle);
		SetIntVec(INT_VEC_ON_KEY_PRESS,OldInt);
		return errCode;
	ENDTRY

	SetIntVec(INT_VEC_ON_KEY_PRESS,OldInt);

	if (kill_search != 1) return -1;

	return 1;
}
