/* Part of the SUDK sudoku puzzle solver/editor */
/* By Alex Wilson, copyright 2006 */

/* Solving algorithm based in part on the freeware work of
   Wim Colgate */

#define SAVE_SCREEN
#define USE_TI92PLUS
#define USE_V200
#define USE_TI89

#define MIN_AMS 200
#define ENABLE_ERROR_RETURN

#include <tigcclib.h>
#include "sudoku.h"
#include "solving.h"
#include "filemenu.h"
#include "opmenu.h"

HSYM sudoku_hsym;
sudoku *pboard;

const char types[] = { OTH_TAG, 0 };
const char *names[] = { "SUDK", "0" };

int selected=0;

void syncHSYM() {
	int i;

	HANDLE newhandle = HeapAlloc(100);
	SYM_ENTRY *file_ptr = DerefSym(sudoku_hsym);
	if (file_ptr->handle != H_NULL) HeapFree(file_ptr->handle);
	MULTI_EXPR *file = HeapDeref((file_ptr->handle = newhandle));

	file->Size = 49;
	file->Expr[0] = 0;

	for (i=0; i<=41; i++) file->Expr[i+1] = pboard->harray[i];

	file->Expr[42] = 0;
	file->Expr[43] = 'S';
	file->Expr[44] = 'U';
	file->Expr[45] = 'D';
	file->Expr[46] = 'K';
	file->Expr[47] = 0;
	file->Expr[48] = OTH_TAG;
}

void syncSudoku() {
	int i;

	SYM_ENTRY *file_ptr = DerefSym(sudoku_hsym);
	MULTI_EXPR *file = HeapDeref(file_ptr->handle);

	for (i=0; i<=41; i++) pboard->harray[i] = file->Expr[i+1];
}

int newSudoku() {
	int i;
	sudoku_hsym = VarNew(types, names);
	if (sudoku_hsym.folder != H_NULL) {
		for (i=0; i<=41; i++) pboard->harray[i] = 0;

		syncHSYM();

		return 1;
	}

	return 0;
}

int openSudoku() {
	int i;

	sudoku_hsym = VarOpen(types, names);
	if (sudoku_hsym.folder != H_NULL) {
		syncSudoku();

		return 1;
	}

	return 0;
}

void redrawBoard() {
	ClrScr();
	drawOutline();
	highlight(selected);
	drawBoard(pboard);

	ST_helpMsg("Press enter for operations");
}

void showSolution() {
	ClrScr();
	drawOutline();
	drawBoard(solution);

	ST_helpMsg("Enter to save, ESC to return");
}

CALLBACK void evhandler(EVENT *ev) {
	int i;
	sudoku *board;
	char msg[40];
	result r;

	if (ev->Type == CM_KEYPRESS) {
		switch (ev->extra.Key.Code) {
			case KEY_QUIT:
			case KEY_ESC:
				syncHSYM();
				ER_throw(1);
				break;

			case '1':
			case '2':
			case '3':
			case '4':
			case '5':
			case '6':
			case '7':
			case '8':
			case '9':
			case '0':
				sudoku_set(pboard,selected, ev->extra.Key.Code - '0');
				highlight(selected);
				drawCell(pboard, selected);
				if (selected < 80)
					selected++;
				highlight(selected);
				break;

			case KEY_RIGHT:
				if (selected < 80) {
					highlight(selected);
					selected++;
					highlight(selected);
				}
				break;
			case KEY_LEFT:
				if (selected > 0) {
					highlight(selected);
					selected--;
					highlight(selected);
				}
				break;
			case KEY_UP:
				if ((selected - 9) > -1) {
					highlight(selected);
					selected -= 9;
					highlight(selected);
				}
				break;
			case KEY_DOWN:
				if ((selected + 9) < 81) {
					highlight(selected);
					selected += 9;
					highlight(selected);
				}
				break;

			case KEY_ENTER:
				opmenu_do();

				if (!strcmp(opmenu, "Verify")) {
					r = check(pboard);
					if (r.failures > 0) {
						ST_helpMsg("Sudoku is not valid.");
					} else {
						if (r.blanks > 0) {
							ST_helpMsg("Sudoku is valid, but has blanks.");
						} else {
							ST_helpMsg("Sudoku is valid.");
						}
					}
				}

				if (!strcmp(opmenu, "Solve")) {
					kill_search = 0;

					r = check(pboard);
					if (r.failures > 0) {
						ST_helpMsg("Puzzle has duplicates, can't solve.");
					} else {

					ST_helpMsg("Please wait while solving...");

					i = solve(pboard);

					if (i==1) {
						showSolution();
						while (i=ngetchx()) {
							if (i == KEY_ENTER) {
								memcpy(pboard, solution, sizeof(sudoku));
								redrawBoard();
								ST_helpMsg("Solution saved.");
								break;
							}
							if (i == KEY_ESC) {
								redrawBoard();
								ST_helpMsg("Solution discarded.");
								break;
							}
						}
					} else {
						if (i==0) {
							ST_helpMsg("Not enough memory to solve.");
						} else {
							if (i==-1) {
								ST_helpMsg("Could not solve puzzle.");
							} else {
								sprintf(msg, "Error #%i occured", i);
								ST_helpMsg(msg);
							}
						}
					}

					HeapFree(sol_handle);
					}
				}

				break;

			case KEY_F1:
				if (filemenu_do()) {
					if (!strcmp(filemenu, "New")) {
						syncHSYM();
						if (!newSudoku()) {
							syncSudoku();
						}
						redrawBoard();
					}
					if (!strcmp(filemenu, "Open")) {
						syncHSYM();
						if (!openSudoku()) {
							syncSudoku();
						}
						redrawBoard();
					}
					if (!strcmp(filemenu, "Quit")) {
						syncHSYM();
						ER_throw(1);
					}
				}
				break;
		}
	}
}

void _main(void) {
	HANDLE hboard = HeapAllocHigh(sizeof(sudoku));
	pboard = HeapDeref(hboard);

	sudoku_hsym = HS_NULL;

	ClrScr();
	drawOutline();

	getfilemenu:

	while (!filemenu_do()) {}

	if (!strcmp(filemenu, "New")) {
		if (!newSudoku()) {
			goto getfilemenu;
		}
	}

	if (!strcmp(filemenu, "Open")) {
		if (!openSudoku()) {
			goto getfilemenu;
		}
	}

	if (strcmp(filemenu, "Quit")) {
		redrawBoard();

		EV_captureEvents(evhandler);
		TRY
			EV_eventLoop();
		ONERR
			EV_captureEvents(NULL);
		ENDTRY
	}

	HeapFree(hboard);
}
